/************************************************************************/
/*
This sample demonstrates 
1. how to control 1R pan and tilt by the point in original fisheye image

	------------------------	   
	|		   |           |
	|     O    |     R     |
	|          |           |
	------------------------	  
*/
/************************************************************************/
#include <iostream>
#include <stdio.h>
#include "fisheyesw.h"
#include "fisheyesw_errordef.h"

#if defined ( _MSC_VER )
#include "SDL.h"
#pragma comment(lib, "SDL2main.lib")
#pragma comment(lib, "SDL2.lib")
#pragma comment(lib, "d_fisheyesw.lib")
#elif defined( __APPLE__ ) && defined( __MACH__ )
#include <SDL2/SDL.h>
#else
#include "SDL.h"
#include <string.h>
#endif

#ifdef main
#undef main
#endif

#define IMAGE_NAME "fisheye_ceiling.bmp"
#define MOUNT_TYPE 1

#define DISPLAY_WINDOW_W 960
#define DISPLAY_WINDOW_H 480

#define DISPLAY_FISHEYE_X 0
#define DISPLAY_FISHEYE_Y 0
#define DISPLAY_FISHEYE_W 480
#define DISPLAY_FISHEYE_H DISPLAY_WINDOW_H

#define DISPLAY_RECTILINEAR_X 480
#define DISPLAY_RECTILINEAR_Y 0
#define DISPLAY_RECTILINEAR_W 480
#define DISPLAY_RECTILINEAR_H DISPLAY_WINDOW_H

#define POLYGON_COUNT 48

typedef struct T_FisheyePTZPosition
{
	float pan;
	float tilt;
	float zoom;
} TFisheyePTZPos;

// keyboard/mouse event
void EventLoop(HANDLE hObject, FEOPTION *pOption, TFisheyePTZPos* pPTZPos, SDL_Texture* pTextureO, SDL_Texture* pTextureR, SDL_Renderer *pRenderer);
SCODE UpdateOutVPicture(HANDLE hObject, FEOPTION *ptOption, SDL_Texture* pTextureO, SDL_Texture* pTextureR, SDL_Renderer *pRenderer);
SCODE DrawDewarpRegion(HANDLE hObject, FEOPTION *ptOption, SDL_Renderer *pRenderer);

int main(int argc, char *argv[])
{
    char *pszimgPath = IMAGE_NAME;
    int mountType = MOUNT_TYPE;

    /************************************************************************/
    /*  SDL instance                                                        */
    /************************************************************************/
    int sdlRet = 0;
    SDL_Texture *pTextureR = NULL;
	SDL_Texture *pTextureO = NULL;
    SDL_Window *pScreen = NULL;
    SDL_Renderer *pRenderer = NULL;
    SDL_Surface* pImage = NULL;
    SDL_Surface* pBitmap = NULL;

    // SDL renderer
    sdlRet = SDL_Init(SDL_INIT_VIDEO);
    if (sdlRet)
    {
        return sdlRet;
    }
    if (NULL == (pScreen = SDL_CreateWindow("dewarp_OcontrolR", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, DISPLAY_WINDOW_W, DISPLAY_WINDOW_H, 0)))
    {
        return -1;
    }
    if (NULL == (pRenderer = SDL_CreateRenderer(pScreen, -1, 0)))
    {
        return -1;
    }
    if (NULL == (pTextureR = SDL_CreateTexture(pRenderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_STREAMING, DISPLAY_RECTILINEAR_W, DISPLAY_RECTILINEAR_H)))
    {
        return -1;
    }
    
    // Load image file
    if (NULL == (pImage = SDL_LoadBMP(pszimgPath)))
    {
        return -1;
    }
    
    // convert image pixel format to RGBA
    SDL_PixelFormat format;
    memset(&format, 0, sizeof(SDL_PixelFormat));
    format.BitsPerPixel = 32;
    format.BytesPerPixel = 4;
    format.Rmask = 0xFF000000; format.Rshift = 0;
    format.Gmask = 0x00FF0000; format.Gshift = 8;
    format.Bmask = 0x0000FF00; format.Bshift = 16;
    format.Amask = 0x000000FF; format.Ashift = 32;
    pBitmap = SDL_ConvertSurface(pImage, &format, 0);
    SDL_FreeSurface(pImage);
    
	if (NULL == (pTextureO = SDL_CreateTextureFromSurface(pRenderer, pBitmap)))
	{
		return -1;
	}

	/************************************************************************/
	/*  Fisheye library instance                                            */
	/************************************************************************/
    SCODE scRet = FISHEYE_S_OK;
    HANDLE hFECtx = NULL;
    TFisheyePTZPos positionPTZ = { 0, 0, 1};
	FEOPTION option;
	memset(&option, 0, sizeof(FEOPTION));

    // 1. Initialize fisheyesw library
    scRet = Fisheye_Initial(&hFECtx, LIBFISHEYE_VERSION);
    if (FISHEYE_S_OK != scRet)
    {
        goto RELEASE_MEMORY;
    }
    
    // 2. Set fisheye options
    // Set input fisheye image information
    option.InVPicture.Width = pBitmap->w;
    option.InVPicture.Height = pBitmap->h;
    option.InVPicture.Stride = pBitmap->pitch;
    option.InVPicture.Format = FE_PIXELFORMAT_RGB32;
    option.InVPicture.Buffer = (BYTE*)pBitmap->pixels;
    
    // Set ouput image information
    option.OutVPicture.Width = DISPLAY_RECTILINEAR_W;
    option.OutVPicture.Height = DISPLAY_RECTILINEAR_H;
    option.OutVPicture.Stride = DISPLAY_RECTILINEAR_W * 4;
    option.OutVPicture.Format = FE_PIXELFORMAT_RGB32;
    option.OutVPicture.Buffer = (BYTE*)malloc(option.OutVPicture.Stride * option.OutVPicture.Height * sizeof(BYTE));
    if (NULL == option.OutVPicture.Buffer)
    {
        goto RELEASE_MEMORY;
    }
    memset(option.OutVPicture.Buffer, 0, option.OutVPicture.Stride * option.OutVPicture.Height * sizeof(BYTE));
    
    // Set FOV information
    option.FOVCenter.X = option.InVPicture.Width >> 1;
    option.FOVCenter.Y = option.InVPicture.Height >> 1;
    option.FOVRadius = option.InVPicture.Height >> 1;
    
    // Set display output region
    option.OutRoi.Left = 0;
    option.OutRoi.Top = 0;
    option.OutRoi.Right = DISPLAY_RECTILINEAR_W;
	option.OutRoi.Bottom = DISPLAY_RECTILINEAR_H;
    
    // Set mount type
    switch (mountType)
    {
        case 0:
            option.MountType = FE_MOUNT_WALL;
            break;
        case 1:
            option.MountType = FE_MOUNT_CEILING;
            break;
        case 2:
            option.MountType = FE_MOUNT_FLOOR;
            break;
        default:
            break;
    }
    
    // Set dewarp type
	option.DewarpType = FE_DEWARP_RECTILINEAR;
    
    // Set flags
    option.Flags = (FE_OPTION_INIMAGEHEADER | FE_OPTION_INIMAGEBUFFER | FE_OPTION_OUTIMAGEHEADER | FE_OPTION_OUTIMAGEBUFFER |
                    FE_OPTION_FOVCENTER | FE_OPTION_FOVRADIUS | FE_OPTION_OUTROI | FE_OPTION_MOUNTTYPE | FE_OPTION_DEWARPTYPE);
    
	scRet = Fisheye_SetOption(hFECtx, &option);
	if (FISHEYE_S_OK != scRet)
	{
		 goto RELEASE_MEMORY;
	}

	Fisheye_SetPanTiltZoom(hFECtx, FE_POSITION_ABSOLUTE, positionPTZ.pan, positionPTZ.tilt, positionPTZ.zoom);

    // 3. Update options in fisheyesw library and paint
    UpdateOutVPicture(hFECtx, &option, pTextureO, pTextureR, pRenderer);
    
    // 4. Wait keyboard and mouse event to perform ePTZ control
    EventLoop(hFECtx, &option, &positionPTZ, pTextureO, pTextureR, pRenderer);
    
RELEASE_MEMORY:
    
    // release buffer memory by user
    if (option.OutVPicture.Buffer)
    {
        free(option.OutVPicture.Buffer);
        option.OutVPicture.Buffer = NULL;
    }
    if (hFECtx)
    {
        // Release Fisheye library.
        Fisheye_Release(&hFECtx);
        hFECtx = NULL;
    }
    
    // release SDL resource
    if (pBitmap)
    {
        SDL_FreeSurface(pBitmap);
        pBitmap = NULL;
    }
    if (pTextureO)
    {
        SDL_DestroyTexture(pTextureO);
        pTextureO = NULL;
    }
	if (pTextureR)
	{
		SDL_DestroyTexture(pTextureR);
		pTextureR = NULL;
	}
    if (pRenderer)
    {
        SDL_DestroyRenderer(pRenderer);
        pRenderer = NULL;
    }
    if (pScreen)
    {
        SDL_DestroyWindow(pScreen);
        pScreen = NULL;
    }
  
    SDL_Quit();
    
    return 0;
}

void EventLoop(HANDLE hObject, FEOPTION *ptOption, TFisheyePTZPos* pPTZPos, SDL_Texture* pTextureO, SDL_Texture* pTextureR, SDL_Renderer *pRenderer)
{
    int bExit = 0;
    SDL_Event evt;
    SDL_MouseButtonEvent beginPos = { 0 };
    int bClick = 0;

    while (!bExit)
    {
        SDL_WaitEvent(&evt);
        
		// exit event loop
		if (SDL_QUIT == evt.type)
		{
			bExit = 1;
			break;
		}

		// get mouse position
		int iMouseX, iMouseY;
		SDL_GetMouseState(&iMouseX, &iMouseY);

		if (iMouseX < DISPLAY_RECTILINEAR_X)
		{
			// left side (fisheye image)
			// click on fisheye image and get pan/tilt
			switch (evt.type)
			{
			case SDL_MOUSEBUTTONDOWN:
				{  
					beginPos = evt.button;  // setup the begin position when mouse click
					bClick = 1;

					// get current PTZ
					Fisheye_GetPanTiltZoom(hObject, &pPTZPos->pan, &pPTZPos->tilt, &pPTZPos->zoom);

					// scale mouse location to InVPicture point 
					int iX = (float)beginPos.x * ptOption->InVPicture.Width / DISPLAY_FISHEYE_W;
					int iY = (float)beginPos.y * ptOption->InVPicture.Height / DISPLAY_FISHEYE_H;
					Fisheye_InVPicturePointToPanTilt(hObject, iX, iY, &pPTZPos->pan ,&pPTZPos->tilt);

					// set new pan and tilt, then repaint
					Fisheye_SetPanTiltZoom(hObject, FE_POSITION_ABSOLUTE, pPTZPos->pan, pPTZPos->tilt, pPTZPos->zoom);
					UpdateOutVPicture(hObject, ptOption, pTextureO, pTextureR, pRenderer);
				}
				break;
			case SDL_MOUSEBUTTONUP:
				bClick = 0;             // disable drag image
				break;
			case SDL_MOUSEMOTION:
				if (bClick)
				{
					SDL_MouseButtonEvent movePos = evt.button;

					// scale mouse location to InVPicture point 
					int iX = (float)movePos.x * ptOption->InVPicture.Width / DISPLAY_FISHEYE_W;
					int iY = (float)movePos.y * ptOption->InVPicture.Height / DISPLAY_FISHEYE_H;
					Fisheye_InVPicturePointToPanTilt(hObject, iX, iY, &pPTZPos->pan ,&pPTZPos->tilt);

					// set PTZ and repaint
					Fisheye_SetPanTiltZoom(hObject, FE_POSITION_ABSOLUTE, pPTZPos->pan, pPTZPos->tilt, pPTZPos->zoom);
					UpdateOutVPicture(hObject, ptOption, pTextureO, pTextureR, pRenderer);
				}	
				break;
			}
		}
		else
		{
			// right side (dewarped image) : ePTZ control
			switch (evt.type)
			{
			case SDL_MOUSEBUTTONDOWN:
				beginPos = evt.button;  // setup the begin position when mouse click
				bClick = 1;
				break;
			case SDL_MOUSEBUTTONUP:
				bClick = 0;
				break;
			case SDL_MOUSEMOTION:
				if (bClick)
				{
					// calculate the difference between begin position and current position
					SDL_MouseButtonEvent movePos = evt.button;
					int iDistance_X = (movePos.x - beginPos.x);
					int iDistance_Y = (movePos.y - beginPos.y);
					Fisheye_SetPanTiltZoom(hObject, FE_POSITION_RELATIVE, -iDistance_X, iDistance_Y, 0);

					// set PTZ and repaint
					UpdateOutVPicture(hObject, ptOption, pTextureO, pTextureR, pRenderer);

					// update begin position to current position
					beginPos = movePos;  
				}
				break;
			case SDL_MOUSEWHEEL:
				Fisheye_SetPanTiltZoom(hObject, FE_POSITION_RELATIVE, 0, 0, evt.wheel.y);
				
				// set PTZ and repaint
				UpdateOutVPicture(hObject, ptOption, pTextureO, pTextureR, pRenderer);
				break;
			default:
				break;
			}
		}
	
	}
}

SCODE UpdateOutVPicture(HANDLE hObject, FEOPTION *ptOption, SDL_Texture* pTextureO, SDL_Texture* pTextureR, SDL_Renderer *pRenderer)
{
    // update dewarped result to OutVPicture.Buffer
    SCODE scRet = Fisheye_OneFrame(hObject);
    if (S_OK != scRet)
    {
		return scRet;
    }

	SDL_UpdateTexture(pTextureR, NULL, ptOption->OutVPicture.Buffer, ptOption->OutVPicture.Stride);	

	// SDL display Textures to Render
	SDL_Rect rectToDrawR = {DISPLAY_RECTILINEAR_X, DISPLAY_RECTILINEAR_Y, DISPLAY_RECTILINEAR_W, DISPLAY_RECTILINEAR_H};
	SDL_Rect rectToDrawO = {DISPLAY_FISHEYE_X, DISPLAY_FISHEYE_Y, DISPLAY_FISHEYE_W, DISPLAY_FISHEYE_H}; 
	SDL_RenderClear(pRenderer);
	SDL_RenderCopy(pRenderer, pTextureR, NULL, &rectToDrawR);
	SDL_RenderCopy(pRenderer, pTextureO, NULL, &rectToDrawO);

	// SDL display the polygon of rectilinear dewarp region on the left side (fisheye image)
	DrawDewarpRegion(hObject, ptOption, pRenderer);

	// Present
    SDL_RenderPresent(pRenderer);
    return S_OK;
}

SCODE DrawDewarpRegion(HANDLE hObject, FEOPTION *ptOption, SDL_Renderer *pRenderer)
{
	// Draw center
	int iX = 0, iY = 0;
	Fisheye_OutRoiPointToInVPicturePoint(hObject, DISPLAY_RECTILINEAR_W / 2, DISPLAY_RECTILINEAR_H / 2, &iX, &iY);
	
	// scale InVPicture point to SDL window display location
	iX = (float)iX * DISPLAY_FISHEYE_W / ptOption->InVPicture.Width;
	iY = (float)iY * DISPLAY_FISHEYE_H / ptOption->InVPicture.Height;

	SDL_SetRenderDrawColor(pRenderer, 255, 0, 0, 0);
	SDL_RenderDrawLine(pRenderer, iX - 5, iY, iX + 5, iY);
	SDL_RenderDrawLine(pRenderer, iX, iY - 5, iX, iY + 5);

	// Draw polygon
	SDL_Point points[POLYGON_COUNT + 1];
	SDL_Point* pInVPoints = points;

	int iPointCnt = POLYGON_COUNT / 4;	
	int iOutW = ptOption->OutRoi.Right - ptOption->OutRoi.Left;
	int iOutH = ptOption->OutRoi.Bottom - ptOption->OutRoi.Top;
	int iIncX = iOutW / iPointCnt;
	int iIncY = iOutH / iPointCnt;

	for (int i = 0; i < iPointCnt; i++)
	{
		int iX = i * iIncX;
		Fisheye_OutRoiPointToInVPicturePoint(hObject, iX, 0, &pInVPoints->x, &pInVPoints->y);
		pInVPoints++;
	}
	for (int i = 0; i < iPointCnt; i++)
	{
		int iY = i * iIncY;
		Fisheye_OutRoiPointToInVPicturePoint(hObject, iOutW, iY, &pInVPoints->x, &pInVPoints->y);
		pInVPoints++;
	}
	for (int i = 0; i < iPointCnt; i++)
	{
		int iX = (iOutW) - i * iIncX;
		Fisheye_OutRoiPointToInVPicturePoint(hObject, iX, iOutH, &pInVPoints->x, &pInVPoints->y);
		pInVPoints++;
	}
	for (int i = 0; i < iPointCnt; i++)
	{
		int iY = (iOutH) - i * iIncY;
		Fisheye_OutRoiPointToInVPicturePoint(hObject, 0, iY, &pInVPoints->x, &pInVPoints->y);
		pInVPoints++;
	}
	
	points[POLYGON_COUNT].x = points[0].x;
	points[POLYGON_COUNT].y = points[0].y;
	
	for (int i = 0; i < POLYGON_COUNT + 1; i++)
	{
		points[i].x = (float)points[i].x * DISPLAY_FISHEYE_W / ptOption->InVPicture.Width;
		points[i].y = (float)points[i].y * DISPLAY_FISHEYE_H / ptOption->InVPicture.Height;
	}

	points[POLYGON_COUNT].x = points[0].x;
	points[POLYGON_COUNT].y = points[0].y;

	SDL_SetRenderDrawColor(pRenderer, 255, 255, 0, 255);
	SDL_RenderDrawLines(pRenderer, points, POLYGON_COUNT + 1);

	return S_OK;
}